/*
 * divfrac: 32 bit division with fractional part
 * Copyright (C) 2005 Christian Wolff
 * 2005-03-14
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 * 
 * The author can be reached at sub-divfrac<at>scara.com, 
 * the project's page is at http://scara.com/divfrac/
 */

#include <stdlib.h>
#include <stdio.h>

/* This funtion divides two unsigned 32 bit integer, 
   and writes the resulting integer quotient, the
   remainder and the first 32 bits of the fractional 
   part into 3 32 bit integer, whose references are 
   passed to the function. Each of the references
   can be NULL, if the result is not desired */
int divfrac(
	unsigned long int numer,   /* enumerator                  */
	unsigned long int denom,   /* denominator                 */
	unsigned long int *pQuot,  /* result of integer division  */
	unsigned long int *pRem,   /* remainder of division       */
	unsigned long int *pFrac)  /* fractional part of division */
{
	unsigned long int quot, frac;
	long int shift;
	
	/* division by zero? */
	if (!denom) return -1;
	
	/* align denom with numer */
	quot = frac = 0;
	shift = 0;
	while ((denom < numer) && ((denom & 0x80000000) == 0)) {
		denom <<= 1;
		shift++;
	}
	
	/* divide numer by denom */
	while (1) {
		if (numer >= denom) {
			numer -= denom;
			quot |= 1;
		}
		if (! shift) break;
		denom >>= 1;
		quot <<= 1;
		shift--;
	}
	
	/* write results and check if fractional part is needed */
	if (pQuot) *pQuot = quot;
	if (pRem) *pRem = numer;
	if (!pFrac) return 0;
	
	/* keep dividing remainder for fractional part */
	while (1) {
		if (numer & 0x80000000) {
			if ( ((numer == (denom >> 1)) && (!(denom & 1))) || (numer > (denom >> 1))) {
				numer -= (denom >> 1);
				frac |= 1;
			}
			numer <<= 1;
			if (denom & 1) numer -= 1;
		} else {
			numer <<= 1;
			if (numer >= denom) {
				numer -= denom;
				frac |= 1;
			}
		}
		if (shift <= -31) break;
		frac <<= 1;
		shift--;
	}
	
	/* write fractional part and return */
	*pFrac = frac;
	return 0;
}

int main(int argc, char *argv[]) 
{
	char *endptr;
	unsigned long int e, d, quot, rem, frac;
	
	/* get command line parameter */
	e = d = 0;
	if (argc >= 2) {
		e = strtoul(argv[1], &endptr, 0);
		if (endptr[0] != '\0') e = 0;
		d = strtoul(argv[2], &endptr, 0);
		if (endptr[0] != '\0') d = 0;
	}
	if (!e || !d) {
		printf("usage: %s <enumerator> <denominator>\n");
		exit(1);
	}
	
	/* call divsion function */
	if (divfrac(e, d, &quot, &rem, &frac)) {
		printf("Error: Division by Zero!\n");
		exit(1);
	}
	
	/* print result */
	printf("%lu / %lu = %lu remainder %lu, 0x%08lX.%08lX\n", e, d, quot, rem, quot, frac);
	
	exit(0);
}
